# -*- coding: utf-8 -*-
__author__ = "Bai chenjia"

import argparse
import gym
import numpy as np
import os
import tensorflow as tf
import time
import json
import baselines.common.tf_util as U
from sklearn.preprocessing import PolynomialFeatures
from baselines import logger
from baselines.common.atari_wrappers_deprecated import wrap_dqn
from deepq.replay_buffer import PrioritizedReplayBuffer
from baselines.common.misc_util import (
    boolean_flag,
    pickle_load,
    pretty_eta,
    relatively_safe_pickle_dump,
    set_global_seeds,
    RunningAvg,
    SimpleMonitor
)
from baselines.common.schedules import LinearSchedule, PiecewiseSchedule
from deepq.model import model
from deepq.build_graph import build_train
import sys


def parse_args():
    parser = argparse.ArgumentParser("ATDC-PER《基于TD-error自适应校正的深度Q学习主动采样方法》")
    #
    parser.add_argument("--env", type=str, default="Robotank", help="name of the game")   # 游戏名称
    parser.add_argument("--eps1", type=float, default=50., help="exploration1")
    parser.add_argument("--eps2", type=float, default=7., help="exploration2")    # 探索因子
    # 偏差模型更新周期
    parser.add_argument("--train-freq", type=int, default=1e5, help="")
    parser.add_argument("--save-freq", type=int, default=1e5, help="save model once every time this many iterations are completed")
    parser.add_argument("--seed", type=int, default=42, help="which seed to use")
    #
    parser.add_argument("--replay-buffer-size", type=int, default=int(1e6), help="replay buffer size")
    parser.add_argument("--lr", type=float, default=1e-4, help="learning rate")
    parser.add_argument("--num-steps", type=int, default=int(2e8), help="total number of steps to run the environment for")   # 原来
    parser.add_argument("--batch-size", type=int, default=32, help="number of transitions to optimize at the same time")
    parser.add_argument("--learning-freq", type=int, default=4, help="number of iterations between every optimization step")
    parser.add_argument("--target-update-freq", type=int, default=40000, help="number of iterations between every target network update")
    #
    boolean_flag(parser, "double-q", default=True, help="whether or not to use double q learning")
    parser.add_argument("--prioritized-alpha", type=float, default=0.6, help="alpha parameter for prioritized replay buffer")
    parser.add_argument("--prioritized-beta0", type=float, default=0.4, help="initial value of beta parameters for prioritized replay")
    parser.add_argument("--prioritized-eps", type=float, default=1e-6, help="eps parameter for prioritized replay buffer")
    return parser.parse_args()


def make_env(game_name):
    """
        对游戏环节进行预处理
        1. 添加 No op, Frame Skip, 变为84*84的输入
        2. 添加 Monitor，对周期的 奖励、周期长度、周期结束时间 进行记录
    """
    env = gym.make(game_name + "NoFrameskip-v4")
    monitored_env = SimpleMonitor(env)
    env = wrap_dqn(monitored_env)
    return env, monitored_env


def maybe_save_model(savedir, state):
    """
        保存Q网络、目标Q网络参数，保存经验池
    """
    if savedir is None:
        return
    start_time = time.time()
    model_dir = "model-{}".format(state["num_iters"])
    # 调用tf.train.Save.save 函数将模型保存到本地
    U.save_state(os.path.join(savedir, model_dir, "saved"))
    # 保存经验池
    relatively_safe_pickle_dump(state, os.path.join(savedir, 'training_state.pkl.zip'), compression=True)
    # 保存训练过程数据
    relatively_safe_pickle_dump(state["monitor_state"], os.path.join(savedir, 'monitor_state.pkl'))
    relatively_safe_pickle_dump(state["iter_dict"], os.path.join(savedir, "iter_dict.pkl"))
    np.save(os.path.join(savedir, "coef.npy"), state["coef"])
    logger.log("Saved model in {} seconds\n".format(time.time() - start_time))  # 保存消耗的时间


if __name__ == '__main__':
    args = parse_args()
    if args.env is "":
        print("未设置游戏参数")
    savedir = os.path.join("savedir_correct", args.env)
    if not os.path.exists(savedir):
        print("存储目录不存在: ", savedir)
        sys.exit(1)

    # 建立仿真环境，仿真环境已添加Monitor
    env, monitored_env = make_env(args.env)
    if args.seed > 0:
        set_global_seeds(args.seed)
        env.unwrapped.seed(args.seed)

    # 保存输入的参数
    if savedir:
        with open(os.path.join(savedir, 'args.json'), 'w') as f:
            json.dump(vars(args), f)

    with U.make_session(4) as sess:
        def model_wrapper(img_in, num_actions, scope, **kwargs):
            actual_model = model
            return actual_model(img_in, num_actions, scope, layer_norm=False, **kwargs)

        return_tdError_bai, act, train, update_target, debug = build_train(
            make_obs_ph=lambda name: U.Uint8Input(env.observation_space.shape, name=name),
            q_func=model_wrapper,
            num_actions=env.action_space.n,
            optimizer=tf.train.AdamOptimizer(learning_rate=args.lr, epsilon=1e-4),
            gamma=0.99,                     # 折扣因子
            grad_norm_clipping=10,          # 梯度规约
            double_q=args.double_q
        )

        approximate_num_iters = args.num_steps / 4
        # 探索因子的变化
        exploration = PiecewiseSchedule([
            (0, 1.0),
            (approximate_num_iters / args.eps1, 0.1),
            (approximate_num_iters / args.eps2, 0.01),
        ], outside_value=0.01)

        # 初始化经验池
        replay_buffer = PrioritizedReplayBuffer(args.replay_buffer_size,
                                                args.prioritized_alpha)
        # Beta
        beta_schedule = LinearSchedule(approximate_num_iters,
                                       initial_p=args.prioritized_beta0,
                                       final_p=1.0)
        U.initialize()
        update_target()
        num_iters = 0
        iter_dict = {}
        coef = np.array([])  # 初始化偏差模型参数
        obs = env.reset()
        num_iters_since_reset = 0
        reset = True
        reward_30_mean = 0       # 保存奖励
        #
        while True:
            num_iters += 1
            num_iters_since_reset += 1
            kwargs = {}
            update_eps = exploration.value(num_iters)                               # 探索因子
            action = act(np.array(obs)[None], update_eps=update_eps, **kwargs)[0]   # 动作选择
            reset = False
            new_obs, rew, done, info = env.step(action)
            replay_buffer.add(obs, action, rew, new_obs, float(done))               # 存储新样本
            obs = new_obs
            if done:
                num_iters_since_reset = 0
                obs = env.reset()
                reset = True

            # 是否训练矫正模型
            if (num_iters >= 1e5) and (num_iters % int(args.train_freq) == 0):
                print("偏差模型更新--------------------------")
                # 存储优先级
                buffer_delta = np.array([replay_buffer._it_sum[i] for i in range(len(replay_buffer._storage))], dtype=np.float)
                buffer_idxs = np.arange(len(replay_buffer._storage)).astype(np.int)
                not_noise_idx = buffer_idxs[buffer_delta != np.max(buffer_delta)].astype(np.int)
                # 存储优先级
                buffer_delta_new = buffer_delta[not_noise_idx]
                buffer_delta_new_max = np.max(buffer_delta_new)
                buffer_delta_prob = buffer_delta_new / buffer_delta_new_max
                # 回放周期
                buffer_time = np.array(replay_buffer.last_sample_bai)[not_noise_idx]
                buffer_time_prob = buffer_time / np.max(buffer_time)
                # 构造多项式特征
                X_train = np.stack([buffer_delta_prob, buffer_time_prob]).transpose()
                poly = PolynomialFeatures(2)    # 多项式特征阶数
                X_train = poly.fit_transform(X_train).astype(np.float)
                # 真实优先级
                td_errors_all = []
                batch_num = int(len(replay_buffer._storage) / 400)
                for i in range(400):
                    experience_all = replay_buffer.sample_all(index=[int(a) for a in range(batch_num*i, batch_num*(i+1))])
                    (s_all, a_all, r_all, s1_all, d_all, index_all) = experience_all
                    # 所有样本送入神经网络，计算TD error，取指数_alpha
                    td_errors_batch = return_tdError_bai(s_all, a_all, r_all, s1_all, d_all)
                    td_errors_all.append(td_errors_batch)
                    experience_all, s_all, a_all, r_all, s1_all, d_all, index_all = [], [], [], [], [], [], []
                td_errors_all = np.hstack(td_errors_all)
                td_errors_all_e = np.abs(td_errors_all) + args.prioritized_eps
                real_delta_prob = np.array([a ** replay_buffer._alpha for a in td_errors_all_e])/buffer_delta_new_max
                # 优先级偏差
                Y_train = real_delta_prob[not_noise_idx] - buffer_delta_prob
                Y_train = Y_train.astype(np.float)
                # 偏差模型求解
                ret = np.linalg.lstsq(X_train, Y_train)
                coef = ret[0]
                print("偏差模型参数为: {}".format(coef))

            # 优先级矫正
            if (num_iters > max(5 * args.batch_size, args.replay_buffer_size // 20) and
                    num_iters % args.learning_freq == 0):
                # 根据模型矫正优先级和训练
                if (num_iters > 1e5) and (coef.shape[0] != 0):
                    buffer_delta = np.array(replay_buffer._it_sum._value[replay_buffer._it_sum._capacity: replay_buffer._it_sum._capacity+len(replay_buffer._storage)], dtype=float)
                    buffer_idxs = np.arange(len(replay_buffer._storage)).astype(int)
                    buffer_delta_correct = buffer_delta.copy()
                    not_noise_idx = buffer_idxs[buffer_delta != np.max(buffer_delta)].astype(int)
                    # 存储优先级
                    buffer_delta_new = buffer_delta[not_noise_idx]
                    buffer_delta_new_max = np.max(buffer_delta_new)
                    buffer_delta_prob = buffer_delta_new / buffer_delta_new_max
                    # 回放周期
                    buffer_time = np.array(replay_buffer.last_sample_bai)[not_noise_idx]
                    buffer_time_prob = buffer_time / np.max(buffer_time)
                    # 多项式特征
                    X_test = np.stack([buffer_delta_prob, buffer_time_prob]).transpose()
                    poly = PolynomialFeatures(2)
                    X_test = poly.fit_transform(X_test)     # 进行多项式变换
                    # 优先级矫正
                    y_pred = np.dot(X_test, coef)
                    # 真实优先级的估计
                    y_pred_delta = (buffer_delta_prob + y_pred) * buffer_delta_new_max
                    #
                    y_pred_delta[y_pred_delta > np.max(buffer_delta)] = np.max(buffer_delta)
                    y_pred_delta[y_pred_delta <= 0.] = args.prioritized_eps
                    buffer_delta_correct[not_noise_idx] = y_pred_delta
                    replay_buffer._it_sum.set_all_new(buffer_delta_correct)
                    replay_buffer._it_min.set_all_new(buffer_delta_correct)
                    # 采样
                    experience = replay_buffer.sample(args.batch_size, beta=beta_schedule.value(num_iters))
                    replay_buffer._it_sum.set_all_new(buffer_delta)
                    replay_buffer._it_min.set_all_new(buffer_delta)
                    (obses_t, actions, rewards, obses_tp1, dones, weights, batch_idxes) = experience
                    # 训练Q网络
                    td_errors = train(obses_t, actions, rewards, obses_tp1, dones, weights)
                    new_priorities = np.abs(td_errors) + args.prioritized_eps
                    # 更新经验池
                    replay_buffer.update_priorities(batch_idxes, new_priorities)

                #
                else:
                    experience = replay_buffer.sample(args.batch_size, beta=beta_schedule.value(num_iters))
                    (obses_t, actions, rewards, obses_tp1, dones, weights, batch_idxes) = experience
                    td_errors = train(obses_t, actions, rewards, obses_tp1, dones, weights)
                    new_priorities = np.abs(td_errors) + args.prioritized_eps
                    replay_buffer.update_priorities(batch_idxes, new_priorities)

            # 更新目标Q网络
            if num_iters % args.target_update_freq == 0:
                update_target()

            # 保存Q网络参数、经验池、训练过程奖励、偏差模型
            if num_iters > 0 and (num_iters % args.save_freq == 0 or info["steps"] > args.num_steps):
                maybe_save_model(savedir, {
                    'replay_buffer': replay_buffer,
                    'num_iters': num_iters,
                    'monitor_state': monitored_env.get_state(),
                    'iter_dict': iter_dict,
                    "coef": coef,
                })
            if info["steps"] > args.num_steps:    # 达到终止时间步
                break
            # 输出日志
            if done:
                reward_30_mean = np.mean(info["rewards"][-100:])
                iter_dict[len(info["rewards"])] = num_iters
                completion = np.round(info["steps"] / args.num_steps, 1)  # 完成率
                logger.record_tabular("% completion", completion)
                logger.record_tabular("steps", info["steps"])
                logger.record_tabular("iters", num_iters)
                logger.record_tabular("episodes", len(info["rewards"]))
                logger.record_tabular("reward (100 epi mean)", np.mean(info["rewards"][-100:]))
                logger.record_tabular("exploration", exploration.value(num_iters))
                logger.record_tabular("reply_storge", len(replay_buffer._storage))
                logger.dump_tabular()
                logger.log()
