# -*- coding: utf-8 -*-

import tensorflow as tf
import baselines.common.tf_util as U


def build_act(make_obs_ph, q_func, num_actions, scope="deepq", reuse=None):
    """
        Creates the act function:
    """
    with tf.variable_scope(scope, reuse=reuse):
        observations_ph = U.ensure_tf_input(make_obs_ph("observation"))

        # stochastic_ph 如果为True则选择 epsilon-贪心动作，否则选择随机动作
        stochastic_ph = tf.placeholder(tf.bool, (), name="stochastic")
        update_eps_ph = tf.placeholder(tf.float32, (), name="update_eps")
        eps = tf.get_variable("eps", (), initializer=tf.constant_initializer(0))

        # 计算所有动作的Q值
        q_values = q_func(observations_ph.get(), num_actions, scope="q_func")

        # 以 epsilon-贪心 选择一个动作
        deterministic_actions = tf.argmax(q_values, axis=1)  # 贪心动作
        batch_size = tf.shape(observations_ph.get())[0]
        random_actions = tf.random_uniform(tf.stack([batch_size]), minval=0, maxval=num_actions, dtype=tf.int64)     # 随机动作
        chose_random = tf.random_uniform(tf.stack([batch_size]), minval=0, maxval=1, dtype=tf.float32) < eps
        # 如果 chose_random 成立，则选择随机动作，否则选择贪心动作
        stochastic_actions = tf.where(chose_random, random_actions, deterministic_actions)

        output_actions = tf.cond(stochastic_ph, lambda: stochastic_actions, lambda: deterministic_actions)
        update_eps_expr = eps.assign(tf.cond(update_eps_ph >= 0, lambda: update_eps_ph, lambda: eps))      # 根据条件选择是否是 epsilon 参数进行更新
        act = U.function(inputs=[observations_ph, stochastic_ph, update_eps_ph],
                         outputs=output_actions,
                         givens={update_eps_ph: -1.0, stochastic_ph: True},
                         updates=[update_eps_expr])
        return act


def build_train(make_obs_ph, q_func, num_actions, optimizer, grad_norm_clipping=None, gamma=1.0,
                double_q=True, scope="deepq", reuse=None):
    # 根据当前观测，调用 build_act 函数选择动作
    act_f = build_act(make_obs_ph, q_func, num_actions, scope=scope, reuse=reuse)
    with tf.variable_scope(scope, reuse=reuse):
        # set up placeholders
        obs_t_input = U.ensure_tf_input(make_obs_ph("obs_t"))           # s
        act_t_ph = tf.placeholder(tf.int32, [None], name="action")      # a
        rew_t_ph = tf.placeholder(tf.float32, [None], name="reward")    # r
        obs_tp1_input = U.ensure_tf_input(make_obs_ph("obs_tp1"))       # s_
        done_mask_ph = tf.placeholder(tf.float32, [None], name="done")  # done  ？？？
        importance_weights_ph = tf.placeholder(tf.float32, [None], name="weight")  # IS

        # Q网络: 输入当前状态、动作数目，输出 batch_size*nA，每一行是一个样本的所有动作的值函数
        q_t = q_func(obs_t_input.get(), num_actions, scope="q_func", reuse=True)
        q_func_vars = U.scope_vars(U.absolute_scope_name("q_func"))

        # target-Q网络：输入下一个状态，动作数目，输出 batch_size*nA，每一行是一个样本的所有动作的值函数
        q_tp1 = q_func(obs_tp1_input.get(), num_actions, scope="target_q_func")
        target_q_func_vars = U.scope_vars(U.absolute_scope_name("target_q_func"))

        # batch_size * 1，每一行为该状态处所选动作的值函数
        q_t_selected = tf.reduce_sum(q_t * tf.one_hot(act_t_ph, num_actions), 1)

        if double_q:
            # Q 网络选择最优动作 q_tp1_best_using_online_net
            q_tp1_using_online_net = q_func(obs_tp1_input.get(), num_actions, scope="q_func", reuse=True)
            q_tp1_best_using_online_net = tf.arg_max(q_tp1_using_online_net, 1)
            # target 网络估计值函数
            q_tp1_best = tf.reduce_sum(q_tp1 * tf.one_hot(q_tp1_best_using_online_net, num_actions), 1)
        else:
            q_tp1_best = tf.reduce_max(q_tp1, 1)
        q_tp1_best_masked = (1.0 - done_mask_ph) * q_tp1_best

        # compute RHS of bellman equation  构造 TD-target
        q_t_selected_target = rew_t_ph + gamma * q_tp1_best_masked

        # 构造 TD-Error，此处使用 tf.stop_gradient 是为了防止target-Q网络的梯度更新
        td_error = q_t_selected - tf.stop_gradient(q_t_selected_target)
        errors = U.huber_loss(td_error)      # 损失
        # 重要性采样权重
        weighted_error = tf.reduce_mean(importance_weights_ph * errors)

        # 选择优化器最小化loss，在if中需要对每个梯度根据norm进行clip（norm不超过10）
        if grad_norm_clipping is not None:
            optimize_expr = U.minimize_and_clip(optimizer,
                                                weighted_error,
                                                var_list=q_func_vars,
                                                clip_val=grad_norm_clipping)
        else:
            optimize_expr = optimizer.minimize(weighted_error, var_list=q_func_vars)

        update_target_expr = []
        for var, var_target in zip(sorted(q_func_vars, key=lambda v: v.name),
                                   sorted(target_q_func_vars, key=lambda v: v.name)):
            update_target_expr.append(var_target.assign(var))
        update_target_expr = tf.group(*update_target_expr)

        train = U.function(
            inputs=[
                obs_t_input,            # s
                act_t_ph,               # a
                rew_t_ph,               # r
                obs_tp1_input,          # s_
                done_mask_ph,           # done
                importance_weights_ph   # IS
            ],
            outputs=td_error,           # td_error
            updates=[optimize_expr]     # train_op
        )
        update_target = U.function([], [], updates=[update_target_expr])

        q_values = U.function([obs_t_input], q_t)

        # 返回所有样本的TD error
        return_tdError_bai = U.function(
            inputs=[
                obs_t_input,            # s
                act_t_ph,               # a
                rew_t_ph,               # r
                obs_tp1_input,          # s_
                done_mask_ph            # done
            ],
            outputs=td_error            # td_error
        )

        # 分别为： 输出epsilon贪心动作，训练Q网络，更新target-Q网络，记录Q网络输出值
        return return_tdError_bai, act_f, train, update_target, {'q_values': q_values}
