# -*- coding: utf-8 -*-

import numpy as np
import random
import operator


class SegmentTree(object):
    def __init__(self, capacity, operation, neutral_element):
        """
            Build a Segment Tree data structure.
            https://en.wikipedia.org/wiki/Segment_tree
        """
        assert capacity > 0 and capacity & (capacity - 1) == 0, "capacity must be positive and a power of 2."
        self._capacity = capacity
        self._value = np.array([neutral_element for _ in range(2 * capacity)], dtype=np.float)
        self._operation = operation

    def _reduce_helper(self, start, end, node, node_start, node_end):
        if start == node_start and end == node_end:
            return self._value[node]
        mid = (node_start + node_end) // 2
        if end <= mid:
            return self._reduce_helper(start, end, 2 * node, node_start, mid)
        else:
            if mid + 1 <= start:
                return self._reduce_helper(start, end, 2 * node + 1, mid + 1, node_end)
            else:
                return self._operation(
                    self._reduce_helper(start, mid, 2 * node, node_start, mid),
                    self._reduce_helper(mid + 1, end, 2 * node + 1, mid + 1, node_end)
                )

    def reduce(self, start=0, end=None):
        """
            Returns result of applying `self.operation`
            to a contiguous subsequence of the array.
            self.operation(arr[start], operation(arr[start+1], operation(... arr[end])))
        """
        if end is None:
            end = self._capacity
        if end < 0:
            end += self._capacity
        end -= 1
        return self._reduce_helper(start, end, 1, 0, self._capacity - 1)

    def __setitem__(self, idx, val):
        # index of the leaf
        idx += self._capacity
        self._value[idx] = val
        idx //= 2
        while idx >= 1:
            self._value[idx] = self._operation(
                self._value[2 * idx],
                self._value[2 * idx + 1]
            )
            idx //= 2

    def __getitem__(self, idx):
        assert 0 <= idx < self._capacity
        return self._value[self._capacity + idx]

    # 新添加函数，用于重置整个sumTree
    def set_all(self, val_list):
        for idx, val in enumerate(val_list):
            idx_real = idx + self._capacity
            self._value[idx_real] = val
        parent_index = self._capacity - 1
        while parent_index >= 1:
            self._value[parent_index] = self._operation(
                self._value[2 * parent_index],
                self._value[2 * parent_index + 1]
            )
            parent_index -= 1

    def print_all(self):
        return self._value
        #for i, val in enumerate(self._value):
        #    print("i={}, val={}".format(i, val))

    def set_all_new(self, val_list):
        #np.save("val_list.npy", np.array(val_list))
        if type(self._value) == list:
            self._value = np.array(self._value, dtype=np.float)
        self._value[self._capacity: self._capacity+len(val_list)] = val_list
        lenth = int(self._capacity / 2)   # 本行的长度
        first = self._capacity

        while(lenth >= 1):
            last = first
            first = last - lenth
            left_index = np.arange(2*first, 2*last)[::2]
            right_index = np.arange(2*first, 2*last)[1::2]
            if self._operation is operator.add:
                self._value[first: last] = np.add(
                    self._value[left_index],
                    self._value[right_index])
            else:
                self._value[first: last] = np.min(np.vstack([
                    self._value[left_index], self._value[right_index]]),
                    axis=0)
            lenth = int(lenth / 2)


class SumSegmentTree(SegmentTree):
    def __init__(self, capacity):
        super(SumSegmentTree, self).__init__(
            capacity=capacity,
            operation=operator.add,
            neutral_element=0.0
        )

    def sum(self, start=0, end=None):
        """Returns arr[start] + ... + arr[end]"""
        return super(SumSegmentTree, self).reduce(start, end)

    def find_prefixsum_idx(self, prefixsum):
        """
            Find the highest index `i` in the array such that
            sum(arr[0] + arr[1] + ... + arr[i - i]) <= prefixsum
        """
        assert 0 <= prefixsum <= self.sum() + 1e-5
        idx = 1
        while idx < self._capacity:  # while non-leaf
            if self._value[2 * idx] > prefixsum:
                idx = 2 * idx
            else:
                prefixsum -= self._value[2 * idx]
                idx = 2 * idx + 1
        return idx - self._capacity

    def print_all(self):
        return super(SumSegmentTree, self).print_all()

    def set_all(self, val_list):
        return super(SumSegmentTree, self).set_all(val_list)

    def set_all_new(self, val_list):
        return super(SumSegmentTree, self).set_all_new(val_list)


class MinSegmentTree(SegmentTree):
    def __init__(self, capacity):
        super(MinSegmentTree, self).__init__(
            capacity=capacity,
            operation=min,
            neutral_element=float('inf')
        )

    def min(self, start=0, end=None):
        """Returns min(arr[start], ...,  arr[end])"""
        return super(MinSegmentTree, self).reduce(start, end)

    def print_all(self):
        return super(MinSegmentTree, self).print_all()

    def set_all(self, val_list):
        return super(MinSegmentTree, self).set_all(val_list)

    def set_all_new(self, val_list):
        return super(MinSegmentTree, self).set_all_new(val_list)


class ReplayBuffer(object):
    def __init__(self, size):
        """
            Create Prioritized Replay buffer.
        """
        self._storage = []
        self._maxsize = size
        self._next_idx = 0

    def __len__(self):
        return len(self._storage)

    def add(self, obs_t, action, reward, obs_tp1, done):
        data = (obs_t, action, reward, obs_tp1, done)

        if self._next_idx >= len(self._storage):
            self._storage.append(data)
        else:
            self._storage[self._next_idx] = data
        self._next_idx = (self._next_idx + 1) % self._maxsize

    def _encode_sample(self, idxes):
        obses_t, actions, rewards, obses_tp1, dones = [], [], [], [], []
        for i in idxes:
            data = self._storage[i]
            obs_t, action, reward, obs_tp1, done = data
            obses_t.append(np.array(obs_t, copy=False))
            actions.append(np.array(action, copy=False))
            rewards.append(reward)
            obses_tp1.append(np.array(obs_tp1, copy=False))
            dones.append(done)
        return np.array(obses_t), np.array(actions), np.array(rewards), np.array(obses_tp1), np.array(dones)

    def sample(self, batch_size):
        """
            Sample a batch of experiences.
        """
        idxes = [random.randint(0, len(self._storage) - 1) for _ in range(batch_size)]
        return self._encode_sample(idxes)


class PrioritizedReplayBuffer(ReplayBuffer):
    def __init__(self, size, alpha):
        """
            Create Prioritized Replay buffer.
        """
        super(PrioritizedReplayBuffer, self).__init__(size)
        assert alpha > 0
        self._alpha = alpha

        # sum_tree的容量必须是2的倍数，该容量即为叶节点个数
        it_capacity = 1
        while it_capacity < size:
            it_capacity *= 2
        self._it_sum = SumSegmentTree(it_capacity)
        self._it_min = MinSegmentTree(it_capacity)
        self._max_priority = 1.0

        #last_sample_bai:该变量样本距离上次取样的时间
        self.last_sample_bai = np.zeros(self._maxsize)

    def add(self, *args, **kwargs):
        """
            新样本加入被赋值为最大的优先级
        """
        idx = self._next_idx
        super().add(*args, **kwargs)
        self._it_sum[idx] = self._max_priority ** self._alpha
        self._it_min[idx] = self._max_priority ** self._alpha
        #新样本，置0
        self.last_sample_bai[idx] = 0

    def _sample_proportional(self, batch_size):
        res = []
        for _ in range(batch_size):
            # TODO(szymon): should we ensure no repeats?
            mass = random.random() * self._it_sum.sum(0, len(self._storage) - 1)
            idx = self._it_sum.find_prefixsum_idx(mass)
            res.append(idx)
        return res

    def sample(self, batch_size, beta):
        """Sample a batch of experiences.
        compared to ReplayBuffer.sample
        it also returns importance weights and idxes
        of sampled experiences.
        """
        assert beta > 0

        idxes = self._sample_proportional(batch_size)
        #print("idxes =", idxes)

        #所有样本的距离上次训练的时间+1，被取样样本置0
        self.last_sample_bai += 1
        self.last_sample_bai[idxes] = 0

        weights = []
        p_min = self._it_min.min() / self._it_sum.sum()
        #print("self._it_min = ", self._it_min.min(), self._it_min._value[1])
        #print("self._it_sum = ", self._it_sum.sum(), self._it_sum._value[1])
        #print("len of self._storage", len(self._storage))
        #print("p_min =", p_min)
        #print("beta =", beta)
        max_weight = (p_min * len(self._storage)) ** (-beta)

        for idx in idxes:
            p_sample = self._it_sum[idx] / self._it_sum.sum()
            weight = (p_sample * len(self._storage)) ** (-beta)
            weights.append(weight / max_weight)
        weights = np.array(weights)
        encoded_sample = self._encode_sample(idxes)
        return tuple(list(encoded_sample) + [weights, idxes])

    def update_priorities(self, idxes, priorities):
        """Update priorities of sampled transitions.

        sets priority of transition at index idxes[i] in buffer
        to priorities[i].

        Parameters
        ----------
        idxes: [int]
            List of idxes of sampled transitions
        priorities: [float]
            List of updated priorities corresponding to
            transitions at the sampled idxes denoted by
            variable `idxes`.
        """
        assert len(idxes) == len(priorities)
        for idx, priority in zip(idxes, priorities):
            assert priority > 0
            assert 0 <= idx < len(self._storage)
            self._it_sum[idx] = priority ** self._alpha
            self._it_min[idx] = priority ** self._alpha

            self._max_priority = max(self._max_priority, priority)

    def save_samples_now(self, filename):
        """
            保存经验池中所有样本的序号和对应的概率
        """
        sample_list = []
        prob_sum = 0
        for index in range(len(self._storage)):
            prob = self._it_sum[index]
            prob_sum += prob
            one_sample = np.hstack((index, prob))  # 一条样本
            sample_list.append(one_sample)

        # 保存
        sample_np = np.stack(sample_list)
        np.savetxt(filename, sample_np, fmt='%d %10.15f')
        print("save priorities to ..", filename)
        #print("prob_sum=", prob_sum, " _it_sum.sum() =", self._it_sum.sum())

    def sample_all(self, index=None):
        if index is None:
            index = [i for i in range(len(self._storage))]
        encoded_sample = self._encode_sample(index)
        return tuple(list(encoded_sample) + [index])




